/*
   Dot3 Bump Mapping.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to perform dot3 bump mapping in DirectX.
   To do this we will create a .x model and bump that in our scene.  To
   perform texture space dot3 bump mapping you need 3 things.  You need
   the vertex normals, s tangents, and binormals.
*/


// Demo: Dot3 Bump Mapping.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObject();
void RenderScene();
void ShutdownDirect3D();


// DirectX objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;

// Textures.
LPDIRECT3DTEXTURE9 Texture = NULL;
LPDIRECT3DTEXTURE9 bumpMap = NULL;

// Vertex shader, pixel shader, vertex decalaration.
LPDIRECT3DVERTEXSHADER9 VertexShader = NULL;
LPDIRECT3DPIXELSHADER9 PixelShader = NULL;
LPDIRECT3DVERTEXDECLARATION9 ShaderDeclaration = NULL;

// Shader declaration.
D3DVERTEXELEMENT9 SDec[] =
{
   {0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
   {0, 12, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL, 0},
	{0, 24, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
   {0, 32, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT, 0},
   {0, 44, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL, 0},
   D3DDECL_END()
};

// Our model object.
LPD3DXMESH Model = NULL;

// Direction of our main light.
D3DXVECTOR4 lightDir = D3DXVECTOR4(0.0f, 0.0f, 1.0f, 0.0f);

// Our main 3 matrices...
D3DXMATRIX WorldMatrix;
D3DXMATRIX ViewMatrix;
D3DXMATRIX ProjectionMatrix;

// Angle is used to rotate the object.  gWidth and gHeight is used for the aspect ratio.
float angle = 0.0;
int gWidth = 0;
int gHeight = 0;


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   int width, height;

   switch(message)
      {
         case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_CLOSE:
         case WM_DESTROY:
            PostQuitMessage(0);
            break;

         case WM_SIZE:
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height == 0) height = 1;
            if(width == 0) width = 1;

            gWidth = width;
            gHeight = height;
            break;

         default:
         break;
      }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // true then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,                                       // The extended style.
                         "UGPClass",                                 // Window class.
                         "Dot3 Bump Mapping - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
      return false;

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = 640;
         Present_Parameters.BackBufferHeight = 480;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                           D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   // Set our states.
   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);
   D3D_Device->SetRenderState(D3DRS_DITHERENABLE, TRUE);
   D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
   D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

   // Lets get the capabilities of the hardware.
   D3D_Device->GetDeviceCaps(&D3DCaps);

   // Check for the correct shader versions.
   if(D3DCaps.VertexShaderVersion < D3DVS_VERSION(2, 0)) return false;
   if(D3DCaps.PixelShaderVersion < D3DPS_VERSION(1, 4)) return false;

   // This will declare the vertex shader declartion (how the shader is set up).
   if(D3D_Device->CreateVertexDeclaration(SDec, &ShaderDeclaration) != D3D_OK)
      return false;

   // This will hold the source of the vertex shader.
	ID3DXBuffer *source = NULL;

   // Load the vertex shader into Direct3D.
	if(D3DXAssembleShaderFromFile("Dot3.vsh", NULL, NULL, NULL, &source, NULL) != D3D_OK)
      return false;

   // Now we create the vertex shader for Direct3D to use.
   if(D3D_Device->CreateVertexShader((DWORD*)source->GetBufferPointer(),&VertexShader) != D3D_OK)
      return false;

   // We can Release this now because we are done.
   if(source != NULL)
      {
         source->Release(); 
         source = NULL; 
      }

   // Load the pixel shader into Direct3D.
	if(D3DXAssembleShaderFromFile("Dot3.psh", NULL, NULL, NULL, &source, NULL) != D3D_OK)
      return false;

   // Now we create the Pixel shader for Direct3D to use.
   if(D3D_Device->CreatePixelShader((DWORD*)source->GetBufferPointer(),&PixelShader) != D3D_OK)
      return false;

   // We can Release this now because we are done.
   if(source != NULL)
      {
         source->Release(); 
         source = NULL; 
      }

   // Initialize out .x model.
   if(!InitializeObject()) return false;

   // Set the projection matrix.
   D3DXMatrixPerspectiveFovLH(&ProjectionMatrix, 45.0f, (float)gWidth/(float)gHeight,
                              0.1f, 1000.0f);
   D3D_Device->SetTransform(D3DTS_PROJECTION, &ProjectionMatrix);

   return true;
}


bool InitializeObject()
{
   // Load the model from the .x file we will bump map.
   if(FAILED(D3DXLoadMeshFromX("sphere.x", D3DXMESH_SYSTEMMEM,
                               D3D_Device, NULL, NULL, NULL, NULL, &Model)))
      {
         MessageBox(NULL, "Error loading Model.x model file...", "Error!", MB_OK);
         return false;
      }

   // Use clone mesh to copy the declaration (SDec) into the model so we can use
   // DirectX functions to calculate the model's normals and tangents.
   Model->CloneMesh(D3DXMESH_MANAGED, SDec, D3D_Device, &Model);

   // Calculate all the normals of the model.
   if(FAILED(D3DXComputeNormals(Model, NULL))) return false;

   // Use D3DXComputeTangent to calculate the tangents of the model.  Send the model,
   // usage index for the tex coords, usage index for the sTangents, usage index
   // for the binormals, true or false if the textures wrap, and a pointer to an array
   // of three DWORDs per face to be filled with adjacent face indices (not used).
   // If you only wanted to calcuate the sTangent but not the binormal, send D3DX_DEFAULT
   // in the binormal usage index parameter.  This function will fail if you don't have
   // 1 or both declared.
   if(FAILED(D3DXComputeTangent(Model, 0, 0, 0, true, NULL)))
      return false;

   // Load the decal image.
   if(D3DXCreateTextureFromFile(D3D_Device, "decal.tga", &Texture) != D3D_OK)
      return false;

   // Load the image to be bump mapped into a temp variable.  We could load in
   // a normal map but instead I will show you how to turn a greyscale image
   // into a normal map using DirectX functions.
   LPDIRECT3DTEXTURE9 temp = NULL;
   if(D3DXCreateTextureFromFile(D3D_Device, "height.tga", &temp) != D3D_OK)
      return false;

   // Create a blank texture for the bumpMap object.
   D3DSURFACE_DESC desc;
   temp->GetLevelDesc(0, &desc);
   
   if(D3DXCreateTexture(D3D_Device, desc.Width, desc.Height, 0, 0, 
                        D3DFMT_A8R8G8B8,D3DPOOL_MANAGED, &bumpMap) != D3D_OK)
      return false;

   // Use D3DXComputeNormalMap to turn an image into a normal map.  We send in the
   // object to store the bump map, the original image, palette, any flags, channel flag
   // specifying the source height, and a scale value.
   D3DXComputeNormalMap(bumpMap, temp, NULL, 0, D3DX_CHANNEL_RED, 30);
   
   // No longer need the original image.
   temp->Release();
   return true;
}


void SetShader()
{
   // This will create a rotation matrix.
   D3DXMatrixRotationY(&WorldMatrix, angle);

   // Add 0.003 to angle to rotate the object even more the next frame.
   angle += 0.003f;
   if(angle >= 360) angle = 0.0f;

   // Create a view matrix (set the camera).
   D3DXVECTOR3 pos(0.0f, 0.0f, -2.5f);
   D3DXVECTOR3 lookat(0.0f, 0.0f, 0.0f);
   D3DXVECTOR3 up(0.0f, 1.0f, 0.0f);
   D3DXMatrixLookAtLH(&ViewMatrix, &pos, &lookat, &up);

   // Set the model view projection matrix.
   D3DXMATRIX modelViewProj = WorldMatrix * ViewMatrix * ProjectionMatrix;

   // Transpose the model view matrix and set it to constant register 0 (c0 - c3).
   D3DXMatrixTranspose(&modelViewProj, &modelViewProj);
   D3D_Device->SetVertexShaderConstantF(0, (float*)modelViewProj, 4);

   // Transpose the world matrix and set it to constant register 4 (c4 - c7).
   D3DXMatrixTranspose(&WorldMatrix, &WorldMatrix);
   D3D_Device->SetVertexShaderConstantF(4, (float*)WorldMatrix, 4);

   // Set the light direction to constant register 8 (c8).
   D3D_Device->SetVertexShaderConstantF(8, (float*)lightDir,  1);

   // Set how the vertex shader is defined.
   D3D_Device->SetVertexDeclaration(ShaderDeclaration);

   // Set the vertex shader we loaded.
   D3D_Device->SetVertexShader(VertexShader);
   
   // Set the pixel shader we loaded.
   D3D_Device->SetPixelShader(PixelShader);
}


void RenderScene()
{
   // Clear the screen to black.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct3D we are now going to start drawing.
   D3D_Device->BeginScene();

      // Apply the textures.
      D3D_Device->SetTexture(0, Texture);
      D3D_Device->SetTexture(1, bumpMap);

      // Set the vertex shader variables.
      SetShader();

      // Draw the model.
      Model->DrawSubset(0);
    
   // We are done drawing for this scene.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects and vertex shader.
   if(Texture != NULL) 
      {
         Texture->Release();
         Texture = NULL;
      }
   
   if(bumpMap != NULL) 
      {
         bumpMap->Release();
         bumpMap = NULL;
      }

   if(Model != NULL)
      {
         Model->Release();
         Model = NULL;
      }

   if(PixelShader != NULL)
      {
         PixelShader->Release();
         PixelShader = NULL;
      }

   if(VertexShader != NULL)
      {
         VertexShader->Release();
         VertexShader = NULL;
      }

   if(ShaderDeclaration != NULL)
      {
         ShaderDeclaration->Release();
         ShaderDeclaration = NULL;
      }

   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }
}


// Recap:

// And that is how you perform dot3 bump mapping.  The object we used was
// a .x model.  We then used various DirectX functions to calculate the normals,
// tangents, and binormals.  Normally you would calculate this yourself if you
// were using something other than a x mesh model.  All of the real work takes
// place in the vertex and pixel shaders.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com